import ast
import os
import uuid
from types import FunctionType, LambdaType
from typing import Dict, List, Optional, Set, Union

import sexpr

from . import nodes

cd = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

grammar = sexpr.load(os.path.join(cd, 'predicate.yml'))
rules = grammar.yaml


def read_argnames(sexp: List) -> Set[str]:
    '''
    Builds expression's argument list. Traversal is performeed
    in a DFS manner.
    '''
    names = set()

    if isinstance(sexp, list) and sexp and sexp[0] in grammar.top_tags:
        if sexp[0] == 'identifier':
            return set([sexp[1]])
        else:
            for child in sexp[1:]:
                names = names.union(read_argnames(child))

    return names


def compile_sexpr(sexp: List, closure: Optional[Dict] = None):
    '''
    Compiles s-expression into AST expresssion. For a list
    of possible types, check out `auk.nodes` module.

    For explanation regarding `closure` see `compile_terminal`.
    '''
    closure = closure or {}
    args = []

    if not (isinstance(sexp, list) and sexp and sexp[0] in grammar.top_tags):
        return compile_terminal(sexp, closure)

    tag = sexp[0]

    # ast.Name expects string, not ast.Str.
    # Therefore, we must stop recursion here.
    if tag == 'identifier':
        args = sexp[1:]
    else:
        for exp in sexp[1:]:
            arg, closure = compile_sexpr(exp, closure)
            args.append(arg)

    return getattr(nodes, 'compile_%s' % tag)(*args), closure


def compile_terminal(sexp: List, closure: Dict):
    '''
    Compiles primitive type variable (language built-in) to AST node.

    `closure` contains name bindings for target function.
    Variables of non-primitive (built-in) types (without corresponding
    AST node) are given a unique random name, converted to ast.Name
    and stored in closure for the target function.
    '''
    if type(sexp) is bool:
        return ast.NameConstant(value = sexp), closure

    elif type(sexp) in (int, float, complex):
        return ast.Num(n = sexp), closure

    elif type(sexp) is str:
        return ast.Str(s = sexp), closure

    elif type(sexp) is bytes:
        return ast.Bytes(s = sexp), closure

    elif type(sexp) is list:
        elts = []
        for e in sexp:
            e, closure = compile_terminal(e, closure)
            elts.append(e)
        return ast.List(elts = elts, ctx = ast.Load()), closure

    elif type(sexp) is tuple:
        elts = []
        for e in sexp:
            e, closure = compile_terminal(e, closure)
            elts.append(e)
        return ast.Tuple(elts = elts, ctx = ast.Load()), closure

    elif type(sexp) is dict:
        keys, values = [], []
        for k, v in sexp.items():
            k, closure = compile_terminal(k, closure)
            v, closure = compile_terminal(v, closure)
            keys.append(k)
            values.append(v)
        return ast.Dict(keys = keys, values = values), closure

    elif type(sexp) is set:
        elts = []
        for e in sexp:
            e, closure = compile_terminal(e, closure)
            elts.append(e)
        return ast.Set(elts=elts), closure

    else:
        # Generate random name and store variable in closure.
        name = '_%s' % uuid.uuid4().hex
        closure[name] = sexp
        return ast.Name(id = name, ctx = ast.Load()), closure


def compile_predicate(sexp: List, funcname: str = None) -> Union[FunctionType, LambdaType]:
    '''
    Compiles s-expression into predicate function.

    S-expression is validated by the grammar (`predicate.yml`) and,
    if ill-formed, `None` is returned.

    By default, s-expressions with one or zero arguments are compiled
    to lambdas and everything else to proper Python function. This is
    in order to allow passing arguments via kwargs instead of regular
    args. Although argument names are always read in a DFS manner,
    with complex s-expressions, keeping track of the argument order is
    very cumbersome. Compilation to either lambda or function can be
    forced with `force_lambda` and `force_function`, respectively.

    If compiling to function, `compile_predicate` can be provided with
    the name. If `funcname` is null, random UUID is generated.
    '''
    if not grammar.matches(sexp):
        return None

    argnames = read_argnames(sexp)
    exp, env = compile_sexpr(sexp, closure = {})
    funcname = funcname or '_%s' % uuid.uuid4().hex
    arg_list = [ast.arg(arg = name, annotation = None) for name in argnames]

    func_def = ast.FunctionDef(
        name = funcname,
        args = ast.arguments(
            args        = arg_list,
            vararg      = None,
            kwonlyargs  = [],
            kw_defaults = [],
            kwarg       = None,
            defaults    = [],
        ),
        body = [ast.Return(value = exp)],
        decorator_list = [],
        returns = None,
    )

    func_def = ast.fix_missing_locations(func_def)
    mod = ast.Module(body = [func_def])
    exec(compile(mod, '<string>', mode = 'exec'), env)
    return env[funcname]



from auk import compile_predicate


def test_example_1():
    sexp = \
        ['not',
            ['and',
                ['tautology'],
                ['eq',
                    ['identifier', 'lucky_number'],
                    777
                ]
            ]
        ]

    func = compile_predicate(sexp)
    assert not func(777)
    assert func(7)


def test_example_2():
    P, Q = True, False

    sexp = \
        ['not',
            ['or',
                ['identifier', 'P'],
                ['identifier', 'Q'],
            ]
        ]

    func = compile_predicate(sexp)
    assert func(P, Q) == (not P and not Q)

def test_example_3():
    # num in [1, 2, 3] and num >= 2

    sexp = \
        ['and',
            ['in', ['identifier', 'num'], [1, 2, 3]],
            ['gte', ['identifier', 'num'], 2]
        ]

    func = compile_predicate(sexp)
    assert not func(1)
    assert func(2)
    assert func(3)




import ast


def compile_comparison(first, second, opnode):
    return ast.Compare(
        ops = [opnode()],
        left = first,
        comparators = [second]
    )

def compile_eq(first, second):
    return compile_comparison(first, second, ast.Eq)

def compile_neq(first, second):
    return compile_comparison(first, second, ast.NotEq)

def compile_lt(first, second):
    return compile_comparison(first, second, ast.Lt)

def compile_lte(first, second):
    return compile_comparison(first, second, ast.LtE)

def compile_gt(first, second):
    return compile_comparison(first, second, ast.Gt)

def compile_gte(first, second):
    return compile_comparison(first, second, ast.GtE)

def compile_in(first, second):
    return compile_comparison(first, second, ast.In)

def compile_identifier(name):
    return ast.Name(id = name, ctx = ast.Load())

def compile_tautology():
    return ast.NameConstant(value = True)

def compile_contradiction():
    return ast.NameConstant(value = False)

def compile_not(operand):
    return ast.UnaryOp(op = ast.Not(), operand = operand)

def compile_and(*values):
    return ast.BoolOp(op = ast.And(), values = list(values))

def compile_or(*values):
    return ast.BoolOp(op = ast.Or(), values = list(values))

def compile_callable(func):
    return ast.Call(func = func, args = [], keywords = [])



from auk.compiler import read_argnames

def test_read_argnames():
    sexp = ['eq', ['identifier', 'bar'], True]
    argnames = read_argnames(sexp)
    assert argnames == set(['bar'])

    sexp = ['eq', ['identifier', 'bar'], ['identifier', 'foo']]
    argnames = read_argnames(sexp)
    assert argnames == set(['bar', 'foo'])

    sexp = \
        ['not',
            ['and',
                ['identifier', 'arg1'],
                ['eq',
                    ['identifier', 'arg2'],
                    ['identifier', 'arg3']
                ]
            ],
            ['not', ['identifier', 'arg4']]
        ]

    argnames = read_argnames(sexp)
    assert argnames == set(['arg1', 'arg2', 'arg3', 'arg4'])



import ast

from auk.nodes import *


def test_compile_and():
    body = [object(), object()]
    tree = compile_and(*body)
    assert type(tree) is ast.BoolOp
    assert type(tree.op) is ast.And
    assert tree.values == body

def test_compile_or():
    body = [object(), object()]
    tree = compile_or(*body)
    assert type(tree) is ast.BoolOp
    assert type(tree.op) is ast.Or
    assert tree.values == body

def test_compile_not():
    operand = object()
    tree = compile_not(operand)
    assert type(tree) is ast.UnaryOp
    assert type(tree.op) is ast.Not
    assert tree.operand is operand

def test_compile_tautology():
    tree = compile_tautology()
    assert type(tree) is ast.NameConstant
    assert tree.value is True

def test_compile_contradiction():
    tree = compile_contradiction()
    assert type(tree) is ast.NameConstant
    assert tree.value is False

def test_compile_identifier():
    name = object()
    tree = compile_identifier(name)
    assert tree.id is name
    assert type(tree.ctx) is ast.Load

def test_comparisons():
    def test_comparison(compile_func, opcls):
        first, second = object(), object()
        tree = compile_func(first, second)
        assert type(tree) is ast.Compare
        assert type(tree.ops[0]) is opcls
        assert tree.left is first
        assert tree.comparators == [second]

    test_comparison(compile_eq, ast.Eq)
    test_comparison(compile_neq, ast.NotEq)
    test_comparison(compile_lt, ast.Lt)
    test_comparison(compile_lte, ast.LtE)
    test_comparison(compile_gte, ast.GtE)
    test_comparison(compile_in, ast.In)



from setuptools import setup

setup(
    name             = 'auk',
    version          = '0.1.0',
    license          = 'MIT',
    requires         = ['python (>= 3.4)'],
    provides         = ['auk'],
    description      = 'Micro-package for compiling s-expressions into Python\'s predicate functions' ,
    url              = 'http://github.com/IwoHerka/auk',
    packages         = ['auk', 'tests'],
    maintainer       = 'Iwo Herka',
    maintainer_email = 'hi@iwoherka.eu',
    author           = 'Iwo Herka',
    author_email     = 'hi@iwoherka.eu',

    include_package_data = True,
    package_data = dict(booldog = ['predicate.yml']),
    install_requires = [
        'sexpr'
    ],

    classifiers  = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python'
    ],
)



import ast
import pytest

from auk.compiler import *


class Foo(object):
    pass


def test_compile_terminal():
    assert type(compile_terminal(True, {})[0]) is ast.NameConstant
    assert type(compile_terminal(1, {})[0]) is ast.Num
    assert type(compile_terminal(1.0, {})[0]) is ast.Num
    assert type(compile_terminal(complex(1, 1), {})[0]) is ast.Num
    assert type(compile_terminal('str', {})[0]) is ast.Str
    assert type(compile_terminal(b'1', {})[0]) is ast.Bytes
    assert type(compile_terminal([], {})[0]) is ast.List
    assert type(compile_terminal(tuple(), {})[0]) is ast.Tuple
    assert type(compile_terminal({}, {})[0]) is ast.Dict
    assert type(compile_terminal(set(), {})[0]) is ast.Set

    scope = {}
    assert type(compile_terminal(object(), scope)[0]) is ast.Name
    assert scope

    scope = {}
    assert type(compile_terminal(Foo(), scope)[0]) is ast.Name
    assert scope


def test_tautology():
    func = compile_predicate(['tautology'])
    assert func()
    with pytest.raises(TypeError):
        func(0)


def test_contradiction():
    func = compile_predicate(['contradiction'])
    assert not func()
    with pytest.raises(TypeError):
        func(0)


def test_object_literals():
    obj = object()
    sexp = ['eq', ['identifier', 'foo'], obj]
    func = compile_predicate(sexp)
    assert func(obj)

    foo = Foo()
    sexp = ['eq', ['identifier', 'bar'], foo]
    func = compile_predicate(sexp)
    assert func(foo)

    obj = [1, 2, 3]
    sexp = ['eq', ['identifier', 'bar'], obj]
    func = compile_predicate(sexp)
    assert func(obj)

    obj = [Foo(), Foo(), Foo()]
    sexp = ['eq', ['identifier', 'bar'], obj]
    func = compile_predicate(sexp)
    assert func(obj)

    obj = set([1, 2, 3])
    sexp = ['eq', ['identifier', 'bar'], obj]
    func = compile_predicate(sexp)
    assert func(obj)

    obj = dict(a = 1, b = 2, c = 3)
    sexp = ['eq', ['identifier', 'bar'], obj]
    func = compile_predicate(sexp)
    assert func(obj)

    obj = (1, 2, 3)
    sexp = ['eq', ['identifier', 'bar'], obj]
    func = compile_predicate(sexp)
    assert func(obj)


def test_in():
    sexp = ['in', ['identifier', 'foo'], [1, 2, 3]]
    func = compile_predicate(sexp)
    assert func(foo = 1)
    assert not func(4)

    sexp = ['in', 1, [1, 2, 3]]
    func = compile_predicate(sexp)
    assert func()

    foo = object()
    sexp = ['in', foo, [foo, 2, 3]]
    func = compile_predicate(sexp)
    assert func()


def test_func():
    sexp = ['callable', lambda: True]
    func = compile_predicate(sexp)
    assert func()

    sexp = lambda: False
    func = compile_predicate(sexp)
    assert func()
